#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>

/* Magic numbers */
/*                   nanoseconds */
#define PAUSE_SHORT			1000000
#define INTERVAL_SHORT			500000000
#define VARIATION_PAUSE_SHORT		800000
#define VARIATION_INTERVAL_SHORT	400000000
#define VARIATION_PAUSE_LONG		100000000
#define VARIATION_INTERVAL_LONG		100000000
/*                   and seconds */
#define PAUSE_LONG			1
#define INTERVAL_LONG			2


extern int errno;
long user, root;


void exit_routing_on (int signo)
{
   setreuid (root,root);
   system ("/sbin/sysctl -w net.ipv4.ip_forward=1");
   setreuid (user,root);

   exit (0);
}

main (int argc, char **argv) {

   struct timespec *pause, *interval;
   int one_shot_on = 0, one_shot_off = 0;
   int force_long = 0, force_short = 0, likely_short;
   int argc_remaining, current_arg;
   int n_disrupt = 0;
   int debug_flag = 0;
   int u;
   int i;

   argc_remaining = argc - 1;
   current_arg = 1;

   while (argc_remaining) {
      if (!strcmp (argv[current_arg], "-on")) {
         one_shot_on = 1;
         current_arg++;
         argc_remaining--;
      }
      else if (!strcmp (argv[current_arg], "-off")) {
         one_shot_off = 1;
         current_arg++;
         argc_remaining--;
      }
      else if (!strcmp (argv[current_arg], "-l")) {
         force_long = 1;
         current_arg++;
         argc_remaining--;
      }
      else if (!strcmp (argv[current_arg], "-s")) {
         force_short = 1;
         current_arg++;
         argc_remaining--;
      }
      else if (!strcmp (argv[current_arg], "-debug")) {
         debug_flag = 1;
         current_arg++;
         argc_remaining--;
      }
      else if (argc_remaining >= 2 && !strcmp(argv[current_arg], "-n")) {
         n_disrupt = atoi(argv[current_arg+1]);
         current_arg += 2;
         argc_remaining -= 2;
      }
      else {
         break;
      }
   }

   if (argc_remaining > 0 ||
      force_long && force_short ||
      (force_long || force_short) && (one_shot_on || one_shot_off) ||
      one_shot_on && one_shot_off ||
      one_shot_on && n_disrupt ||
      n_disrupt && one_shot_off ||
      n_disrupt > 100)
   {
      printf ("forwarding: usage: forwarding [-on|-off|[-n <num_disrupt (<101)>][-l|-s]]\n");
      return;
   }

   user = getuid ();
   root = geteuid ();

   /* This is no longer necessary - we are running on the real router.
   system ("/sbin/sysctl -w net.ipv4.conf.eth0.send_redirects=0");
   system ("/sbin/sysctl -w net.ipv4.conf.lo.send_redirects=0");
   system ("/sbin/sysctl -w net.ipv4.conf.default.send_redirects=0");
   system ("/sbin/sysctl -w net.ipv4.conf.all.send_redirects=0");
   */

   if (one_shot_on) {
      setreuid (root,root);
      system ("/sbin/sysctl -w net.ipv4.ip_forward=1");
      exit (0);
   }
   if (one_shot_off) {
      setreuid (root,root);
      system ("/sbin/sysctl -w net.ipv4.ip_forward=0");
      exit (0);
   }

   (void) signal (SIGTERM, exit_routing_on);
   (void) signal (SIGINT, exit_routing_on);
   (void) signal (SIGHUP, exit_routing_on);

   srandom (getpid ());

   if (n_disrupt) {
      pause = (struct timespec *) malloc (sizeof (struct timespec) * n_disrupt);
      interval = (struct timespec *) malloc (sizeof (struct timespec) * n_disrupt);
      likely_short = 0;

      for (i=0; i<n_disrupt; i++) {
         if (force_long || force_short) {
            u = force_short;
         }
         else {
            u = (int) (likely_short * 1.0 * random () / (RAND_MAX + 1.0));
            likely_short += 4;
         }

         if (u) {
            /* short pause, initially small chance but increasing */
            pause[i].tv_sec = 0;
            pause[i].tv_nsec = PAUSE_SHORT;
            u = (int) (100.0 * random () / (RAND_MAX + 1.0));
            pause[i].tv_nsec += (u - 50) * VARIATION_PAUSE_SHORT / 100.0;
            interval[i].tv_sec = 0;
            interval[i].tv_nsec = INTERVAL_SHORT;
            u = (int) (100.0 * random () / (RAND_MAX + 1.0));
            interval[i].tv_nsec += (u - 50) * VARIATION_INTERVAL_SHORT / 100.0;

            if (debug_flag) {
               printf ("short pause %d interval %d\n", pause[i].tv_nsec, interval[i].tv_nsec);
            }
         }
         else {
            /* long pause , initially large chance but decreasing */
            pause[i].tv_sec = PAUSE_LONG;
            u = (int) (100.0 * random () / (RAND_MAX + 1.0));
            pause[i].tv_nsec = (u - 50) * VARIATION_PAUSE_LONG / 100.0;
            if (pause[i].tv_nsec < 0) {
               pause[i].tv_nsec += 1000000000;
               pause[i].tv_sec -= 1;
            }
            interval[i].tv_sec = INTERVAL_LONG;
            u = (int) (100.0 * random () / (RAND_MAX + 1.0));
            interval[i].tv_nsec = (u - 50) * VARIATION_INTERVAL_LONG / 100.0;
            if (interval[i].tv_nsec < 0) {
               interval[i].tv_nsec += 1000000000;
               interval[i].tv_sec -= 1;
            }

            if (debug_flag) {
               printf ("long pause %d interval %d\n", pause[i].tv_nsec, interval[i].tv_nsec);
            }
         }
      }

      if (debug_flag) exit (0);

      for (i=0; i<n_disrupt; i++) {
         setreuid (root,root);
         system ("/sbin/sysctl -w net.ipv4.ip_forward=0 > /dev/null");
         if (nanosleep ((pause+i), NULL) == -1) {
            printf ("ERROR attempting to suspend for precise time: %s: Abort.\n", strerror (errno));
            exit (-1);
         }
         system ("/sbin/sysctl -w net.ipv4.ip_forward=1 > /dev/null");
         setreuid (user,root);
         nanosleep ((interval+i), NULL);
      }
   }
   else {
      pause = (struct timespec *) malloc (sizeof (struct timespec));
      interval = (struct timespec *) malloc (sizeof (struct timespec));

      for (;;) {
         /* real time pauses, no variations to be calculated, all short */
         /* for ITL change to all long */
         pause->tv_sec = 0;
         pause->tv_nsec = PAUSE_LONG;
         interval->tv_sec = 0;
         interval->tv_nsec = INTERVAL_LONG;

         printf ("routing disrupted\n");
         setreuid (root,root);
         system ("/sbin/sysctl -w net.ipv4.ip_forward=0 > /dev/null");
         if (nanosleep (pause, NULL) == -1) {
            printf ("ERROR attempting to suspend for precise time: %s: Abort.\n", strerror (errno));
            exit (-1);
         }
         system ("/sbin/sysctl -w net.ipv4.ip_forward=1 > /dev/null");
         setreuid (user,root);
         printf ("routing restored\n");
         /* for ITL increasing gap by 5 seconds */
         sleep (5);
         nanosleep (interval, NULL);
      }
   }
}

